#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "next_move.h"
#include "symbol.h"

#define MIN_VAL -999999

int main(int argc, char *argv[])
{
  /**************************************************************************************
  ** Utilisation:
    - argv[1] : nombres de duels
    - argv[2] : nombre d'entrees du reseau (module 1)
    - argv[3] : mode de recompense du module 1 (0: normal ou 1: puni pour les nulles)
    - argv[4] : style du module adverse (pour match ordi vs. aleatoire), ou:
    - argv[4] : nombre d'entrees du reseau (module 2)
    - argv[5] : mode de recompense du module 2 (0: normal ou 1: puni pour les nulles)
    en mode humain/ordi '#' permet d'arreter le jeu
  ***************************************************************************************/

  /* regles du jeu: nb_choix actions possibles, matrice "gain de l'ordinateur": (choix 0, 1 ..etc)
  **  a   x   x  .... (gauche: ordi; haut: adversaire)
  **  x   a   x
  **  x   x   x
  **  ..
  **  ..
  ** on demarre a zero, but: avoir le plus possible

  ** Modelisation par un reseau de neurones a n entrees (interet: faire varier n)
  */

  int i,j,k,t;

  /* nombre d'actions possibles dans le jeu (PRS => 3) */
  int nb_choix=3;
  int nb_choix2=nb_choix*nb_choix;

  /* l'etat courant gain/perte a maximiser */
  int etat=0;

  /* sorties pouvant etre selectionnees */
  int *sel_outputs=(int*)malloc(nb_choix*sizeof(int));

  /* ensemble des sorties admissibles, non indexees par nb_choix */
  int* set_outputs,nb_outputs;

  /* sortie choisie et sa valeur */
  int output=0;
  float val,val_output;

  /* nombre de passages dans la boucle */
  int nb_boucles=atoi(argv[1]);

  /* nombre d'entrees du reseau */
  int n=atoi(argv[2]),n2=0;

  int modeRecomp=argv[3][0]-'0';
  int modeRecomp2=0;

  /* mode de jeu: 0 pour vs ordi alea, 1 pour vs module similaire, 2 pour vs humain */
  int mode_jeu=0;
  if (argc==4) mode_jeu=2;
  else if (argc==6) mode_jeu=1;

  int param_ordi=-1; /* definit le style de jeu de l'ordinateur adverse;
                     * -1: aleatoire;
                     * 0: preference choix 0, le reste equiprobable;
                     * 1: preference choix 1, le reste equiprobable;
                     * ...etc */
  if (mode_jeu==0) param_ordi=atoi(argv[4]);

  if (mode_jeu==1) {
    n2=atoi(argv[4]); /* nombre de couches du second module */
    modeRecomp2=argv[5][0]-'0'; /* mode de recompense du second module */
  }

  /* coup de l'adversaire dans {0,1,...,nb_choix} */
  int coup_adversaire=0;

  /* coup de l'adversaire humain */
  char ch_coup_adverse;

  /* historiques de l'adversaire */
  int *hist_adversaire,*hist_adversaire2=NULL;

  /* poids et poids moyens */
  float *poids,*poids2=NULL,*poids_moyens,*poids_moyens2=NULL;

  /* somme des poids */
  float sum,sum2;

  /* la matrice des coefficients du jeu;
   * matrice[i*nb_choix+j]==coeff ligne i colonne j
   * ligne 0 en haut; colonne 0 a gauche */
  int* matrice=(int*)malloc(nb_choix2*sizeof(int));

  /* on demarre avec un PPS classique ;
  sinon: a initialiser avec les valeurs voulues, en fonction de la taille */
  matrice[0]=0;  matrice[1]=1;  matrice[2]=-1;
  matrice[3]=-1;  matrice[4]=0;  matrice[5]=1;
  matrice[6]=1;  matrice[7]=-1;  matrice[8]=0;

  /* initialisations des poids et historiques */
  poids=(float*)calloc(nb_choix2*n,sizeof(float));
  poids_moyens=(float*)calloc(nb_choix2*n,sizeof(float));
  hist_adversaire=(int*)calloc(n,sizeof(int));
  if (mode_jeu==1) {
    poids2=(float*)calloc(nb_choix2*n2,sizeof(float));
    poids_moyens2=(float*)calloc(nb_choix2*n2,sizeof(float));
    hist_adversaire2=(int*)calloc(n2,sizeof(int));
  }

  /* initialisation du generateur aleatoire */
  srand(time(NULL));

  /* boucle principale de jeu */
  for (t=nb_boucles-1; t>=0; t--) {

    /********************************************************************
    ** determination de notre coup: on calcule chaque sortie (module 1)
    *********************************************************************/

    val_output=MIN_VAL;
    nb_outputs=0;
    for (i=nb_choix-1; i>=0; i--) {
      val=0;
      for (j=n-1; j>=0; j--) val+=poids[j*nb_choix2+i*nb_choix+hist_adversaire[j]];
      if (val>val_output) {
        val_output=val;
        for (k=nb_choix-1; k>=0; k--) sel_outputs[k]=-1;
        sel_outputs[i]=1; /* seule cette sortie est a choisir pour l'instant */
        nb_outputs=1;
      }
      else if (val==val_output) { /* ne pas oublier les ex-aequo */
        sel_outputs[i]=1;
        nb_outputs++;
      }
    }

    set_outputs=(int*)malloc(nb_outputs*sizeof(int));
    j=0;
    for (i=nb_choix-1; i>=0; i--)
      if (sel_outputs[i]>=0) {
        set_outputs[j]=i;
        j++;
      }
    output=set_outputs[rand()%nb_outputs];
    free(set_outputs);

    if (mode_jeu==1) {

      /*
      ** determination du coup du second module (pris comme adversaire)
      */

      val_output=MIN_VAL;
      nb_outputs=0;
      for (i=nb_choix-1; i>=0; i--) {
        val=0;
        for (j=n2-1; j>=0; j--) val+=poids2[j*nb_choix2+i*nb_choix+hist_adversaire2[j]];
        if (val>val_output) {
          val_output=val;
          for (k=nb_choix-1; k>=0; k--) sel_outputs[k]=-1;
          sel_outputs[i]=1; /* seule cette sortie est a choisir pour l'instant */
          nb_outputs=1;
        }
        else if (val==val_output) { /* ne pas oublier les ex-aequo */
          sel_outputs[i]=1;
          nb_outputs++;
        }
      }

      set_outputs=(int*)malloc(nb_outputs*sizeof(int));
      j=0;
      for (i=nb_choix-1; i>=0; i--)
        if (sel_outputs[i]>=0) {
          set_outputs[j]=i;
          j++;
        }
      coup_adversaire=set_outputs[rand()%nb_outputs];
      free(set_outputs);
    }

    else {

      /*
      ** determination du coup adverse (inconnu pour le bot)
      */

      if (mode_jeu==0) {
        /* genere le coup adverse a l'aide de 2 nombres aleatoires */
        coup_adversaire=next_computer_move(nb_choix,param_ordi,rand(),rand());
      }
      else {
        printf("'p,P': Paper, 'r,R': Rock or 's,S': Scissors ?\n");
        ch_coup_adverse = getc(stdin); getc(stdin); //hack: eat newline too...
        if (ch_coup_adverse=='#') break;
        coup_adversaire=symbToInt(ch_coup_adverse);
      }
    }


    /*****************************************************************
    ** le resultat du jeu est devoile: on module les poids en fonction
    ******************************************************************/

    etat+=matrice[output*nb_choix+coup_adversaire];
    /* suivant le mode de recompense choisi: (0: normal; 1: puni pour les nulles) */
    if (modeRecomp) {
      j=0;
      for (i=nb_choix2*n-1; i>=0; i--)
        /* si la liaison est en connection avec la sortie choisie.. */
        if ((i%nb_choix2)/nb_choix==output && i%nb_choix==hist_adversaire[i/nb_choix2]) {
          if (matrice[output*nb_choix+coup_adversaire]<=0) poids[i]-=1;
          else poids[i]+=1;
        }
    }
    else {
      if (matrice[output*nb_choix+coup_adversaire]!=0) {
        for (i=nb_choix2*n-1; i>=0; i--)
          /* si la liaison est en connection avec la sortie choisie.. */
          if ((i%nb_choix2)/nb_choix==output && i%nb_choix==hist_adversaire[i/nb_choix2]) {
            if (matrice[output*nb_choix+coup_adversaire]<0) poids[i]-=1;
            else if (matrice[output*nb_choix+coup_adversaire]>0) poids[i]+=1;
          }
      }
    }

    /* reequilibrage des poids */
    sum=0;
    for (i=nb_choix2*n-1; i>=0; i--) sum+=poids[i];
    sum/=(nb_choix2*n);
    for (i=nb_choix2*n-1; i>=0; i--) poids[i]-=sum;

    /* shift des coups precedents du joueur adverse */
    for (i=n-1; i>=1; i--) hist_adversaire[i]=hist_adversaire[i-1];
    hist_adversaire[0]=coup_adversaire;

    /* prise en compte des nouveaux poids */
    for (i=nb_choix2*n-1; i>=0; i--) poids_moyens[i]+=poids[i];

    if (mode_jeu==1) { /* si necessaire on repete l'operation pour le second bot */

      /* suivant le mode de recompense choisi: (0: normal; 1: puni pour les nulles) */
      if (modeRecomp2) {
        j=0;
        for (i=nb_choix2*n2-1; i>=0; i--)
          /* si la liaison est en connection avec la sortie choisie.. */
          if ((i%nb_choix2)/nb_choix==coup_adversaire && i%nb_choix==hist_adversaire2[i/nb_choix2]) {
            if (matrice[output*nb_choix+coup_adversaire]>=0) poids2[i]-=1;
            else poids2[i]+=1;
          }
      }
      else {
        if (matrice[output*nb_choix+coup_adversaire]!=0) {
          for (i=nb_choix2*n2-1; i>=0; i--)
            /* si la liaison est en connection avec la sortie choisie.. */
            if ((i%nb_choix2)/nb_choix==coup_adversaire && i%nb_choix==hist_adversaire2[i/nb_choix2]) {
              if (matrice[output*nb_choix+coup_adversaire]>0) poids2[i]-=1;
              else if (matrice[output*nb_choix+coup_adversaire]<0) poids2[i]+=1;
            }
        }
      }

      /* reequilibrage des poids */
      sum2=0;
      for (i=nb_choix2*n-1; i>=0; i--) sum2+=poids2[i];
      sum2/=(nb_choix2*n);
      for (i=nb_choix2*n-1; i>=0; i--) poids2[i]-=sum2;

      /* shift des coups precedents du joueur adverse */
      for (i=n2-1; i>=1; i--) hist_adversaire2[i]=hist_adversaire2[i-1];
      hist_adversaire2[0]=output;

      /* prise en compte des nouveaux poids */
      for (i=nb_choix2*n2-1; i>=0; i--) poids_moyens2[i]+=poids2[i];
    }


    /* affichage de l'etat en cours */
    if (mode_jeu==2) {
      printf("gains de l'ordinateur: %i\n",etat);
      /* affichage du duel si adversaire humain */
      printf("jeux humain-ordi: %c, %c\n",intToSymb(coup_adversaire),intToSymb(output));
    }
    else if (nb_boucles<100 || (nb_boucles>=100 && t%(nb_boucles/100)==0)) printf("etat actuel: %i\n",etat);

  } /* fin de la boucle principale */

  /* affichage des poids moyens: choix joueur de haut en bas,
  anciens coups adversaire de gauche a droite */
  for (i=0; i<n; i++) {
    printf("\npoids moyens pour l'entree %i :\n",i+1);
    for (j=0; j<nb_choix; j++) {
      for (k=0; k<nb_choix; k++)
        printf("%.2f,   ",poids_moyens[i*nb_choix2+j*nb_choix+k]/nb_boucles);
      printf("\n");
    }
  }

  if (mode_jeu==1) {
    printf("\n***************************************\n");
    for (i=0; i<n2; i++) {
      printf("\npoids moyens adverses de l'entree %i :\n",i+1);
      for (j=0; j<nb_choix; j++) {
        for (k=0; k<nb_choix; k++)
          printf("%.2f,   ",poids_moyens2[i*nb_choix2+j*nb_choix+k]/nb_boucles);
        printf("\n");
      }
    }
  }

  /* liberations memoire */
  free(sel_outputs);
  free(hist_adversaire);
  free(poids);
  free(poids_moyens);
  if (poids2!=NULL) {
    free(hist_adversaire2);
    free(poids2);
    free(poids_moyens2);
  }
  free(matrice);

  return 0;
}
